#!/nfs/soar5/safsoar/7.0beta/soar-7.0.0.beta/bin/sun4/soartk -f
#!/home/schwamb/tcl/soar-7.0.0.beta/soartk -f

# This script implements a GUI for a Soar application.

#--------------------------------------------------------------------------

# Source File INCLUDEs:

# Include the code for the file selection dialog
global soar_library
source $soar_library/FSBox.tcl

#--------------------------------------------------------------------------

# GLOBAL Variables:

# The following variables are listed in alphabetical order.

# The global variable below keeps track of the interpreter that we're
# sending commands to.

global app
set app "$interp_name"

# The global variable below keeps track of whether we're in the
# middle of executing a command entered via the text widget.

global executing
set executing 0

# The fileWindowCounter keeps track of how many files have been
# shown to the user so that a unique window name can be used
# for each new file.

global fileWindowCounter
set fileWindowCounter 0

# The inputString variable contains the characters entered by the
# user for the current command.

global inputString
set inputString ""

# The global variable below keeps track of the last command executed,
# so it can be re-executed in response to !! commands.

global lastCommand
set lastCommand ""

# The "logging" variable keeps track of whether we are logging to
# a file or not.

global logging
set logging 0

# Set the World Wide Web viewer to "mosaic-bg" a Bourne shell
# script file that provides remote control of Mosaic.  If you
# do not have Mosaic, then you can substitute the name of your
# Web viewer here.  If your version of Mosaic is not called 
# "mosaic", then you can change the "mosaic-bg" script to match
# the name used in your environment.

global wwwViewer
set wwwViewer mosaic-bg

#--------------------------------------------------------------------------

# GUI Definition

# Create menus that will be attached to menu bar.  We start by defining
# a frame to contain all the menu buttons.

frame .menu -relief raised -bd 2
pack .menu -side top -fill x

# The following menus are listed in the order they appear in the menu
# bar (except for the "Demos" menu since it is packed to the right
# instead of the left).  Within menu sections, commands are listed
# in the order of the menu entries.  Cascades and sub-cascades are
# listed separately.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# FILE menu

# Menu button

menubutton .menu.file -text "File" -menu .menu.file.m

# Menu 

menu .menu.file.m
.menu.file.m add command -label "Source ..." \
                         -command {ProcessFile {source $file}}
.menu.file.m add separator
.menu.file.m add command -label "Save Production Memory ..." \
                         -command {ProcessFile {rete-net -save $file}}
.menu.file.m add command -label "Load Production Memory ..." \
                         -command {ProcessFile {rete-net -load $file}}
.menu.file.m add separator
.menu.file.m add checkbutton -label "Log" -variable logging \
                             -command LogAction

# Pack menu in menu frame.

pack .menu.file -side left -padx 5

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# AGENT menu

menubutton .menu.agent -text "Agent" -menu .menu.agent.m

menu .menu.agent.m
.menu.agent.m add command -label "Set Go parameters ..." -command NotYet

# The cascade menu of agent names is created dynamically.

menu .menu.agent.m.agents -postcommand {FillAgentsMenu SelectAgent $app} \
                          -tearoff no

pack .menu.agent -side left -padx 5

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# SHOW menu

menubutton .menu.show -text "Show" -menu .menu.show.m

menu .menu.show.m
.menu.show.m add command -label "Chunk Free Problem Spaces" \
                           -command {SendToSoarShell \
                                               "chunk-free-problem-spaces"} 
.menu.show.m add command -label "Current Directory" \
                           -command {SendToSoarShell "pwd"}
.menu.show.m add command -label "Default Print Depth" \
                           -command {SendToSoarShell "default-print-depth"}
.menu.show.m add command -label "Firing Counts" \
                           -command {SendToSoarShell "firing-counts"}
.menu.show.m add command -label "Goal Stack" \
                           -command {SendToSoarShell "print -stack"}
.menu.show.m add command -label "Goal Stack Operators" \
                           -command {SendToSoarShell "print -stack -operators"}
.menu.show.m add cascade -label "List" -menu .menu.show.m.list
.menu.show.m add command -label "Memory" \
                           -command {SendToSoarShell "stats -memory"}
.menu.show.m add command -label "Print" -command NotYet
.menu.show.m add command -label "Rete" \
                           -command {SendToSoarShell "stats -rete"}
.menu.show.m add command -label "Stats" \
                           -command {SendToSoarShell "stats"}

# Cascade menu under the "List" item above.

menu .menu.show.m.list
.menu.show.m.list add command -label "Chunks" \
                                 -command NotYet
#                                -command {SendToSoarShell "list-chunks"}
.menu.show.m.list add command -label "Justifications" \
                                 -command NotYet
#                                -command {SendToSoarShell \
#                                                    "list-justifications"}
.menu.show.m.list add command -label "Productions" \
                                -command {SendToSoarShell "list-productions"}
.menu.show.m.list add separator
.menu.show.m.list add command -label "Commands" \
                                -command {SendToSoarShell "info commands"}
.menu.show.m.list add command -label "Procedures" \
                                -command {SendToSoarShell "info procs"}
.menu.show.m.list add command -label "Variables" \
                                -command {SendToSoarShell "info globals"}

pack .menu.show -side left -padx 5

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# EXCISE menu

menubutton .menu.excise -text "Excise" -menu .menu.excise.m

menu .menu.excise.m
.menu.excise.m add command -label "All" \
                           -command {SendToSoarShell "excise-all"}
.menu.excise.m add command -label "Chunks" \
                           -command {SendToSoarShell "excise-chunks"}
.menu.excise.m add command -label "Task" \
                           -command {SendToSoarShell "excise-task"}

pack .menu.excise -side left -padx 5

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# EXECUTE menu

menubutton .menu.execute -text "Execute" -menu .menu.execute.m

menu .menu.execute.m
.menu.execute.m add command -label "Initialize" \
                        -command {SendToSoarShell "init-soar"}
.menu.execute.m add command -label "Reset" \
                        -command {SendToSoarShell "reset"}
.menu.execute.m add separator
.menu.execute.m add cascade -label "Decision" -menu .menu.execute.m.d
.menu.execute.m add cascade -label "Go" -menu .menu.execute.m.g
.menu.execute.m add command -label "Schedule" \
                        -command {SendToSoarShell "schedule"}
.menu.execute.m add command -label "Stop" \
                        -command {SendToSoarShell "stop-soar"}

# Cascade menu under the "Decision" item above

menu .menu.execute.m.d
.menu.execute.m.d add command -label "1" \
                          -command {SendToSoarShell "d 1"}
.menu.execute.m.d add command -label "forever" \
                          -command {SendToSoarShell "d"}

# Cascade menu under the "Decision" item above

menu .menu.execute.m.g
.menu.execute.m.g add command -label "Proceed" \
                          -command {SendToSoarShell "go"}
.menu.execute.m.g add command -label "Set Parameters" -command {NotYet}

pack .menu.execute -side left -padx 5

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# HELP menu

menubutton .menu.help -text "Help" -menu .menu.help.m

menu .menu.help.m
.menu.help.m add cascade -label "On Soar" -menu .menu.help.m.soar
.menu.help.m add cascade -label "On Tcl" -menu .menu.help.m.tcl
.menu.help.m add cascade -label "On Tk" -menu .menu.help.m.tk

# Cascade menu for the "On Soar" item above.

menu .menu.help.m.soar
.menu.help.m.soar add command -label "Change Log" \
                   -command {ShowFile $soar_library/../doc/ChangeLog}
.menu.help.m.soar add command -label "Documentation" \
	           -command {ShowFile $soar_library/../DOCUMENTATION}
.menu.help.m.soar add command -label "Init File" \
                          -command {ShowFile $soar_library/soar.tcl}
.menu.help.m.soar add command -label "Manual Pages" \
                          -command {ShowManPage $soar_library/../doc/cat 0}
.menu.help.m.soar add command -label "News" \
                              -command {SendToSoarShell "soarnews"}
.menu.help.m.soar add command -label "Version" \
                              -command {SendToSoarShell "version"}

.menu.help.m.soar add separator
.menu.help.m.soar add cascade -label "WWW" -menu .menu.help.m.soar.www

# Subcascade for "On Soar/WWW"

menu .menu.help.m.soar.www
.menu.help.m.soar.www add command -label "Soar" \
              -command "exec $wwwViewer http://www.isi.edu/soar/soar.html &"
.menu.help.m.soar.www add command -label "Soar Archive" \
  -command "exec $wwwViewer http://www.isi.edu/soar/soar-archive-homepage.html &"
.menu.help.m.soar.www add command -label "Soar User's Manual" \
  -command "exec $wwwViewer http://www.isi.edu/soar/users-manual/html/soar6-manual.info.Top.html &"
.menu.help.m.soar.www add command -label "Manual Pages" \
 -command "exec $wwwViewer file:$soar_library/../doc/html/soar_man.html &"

# Cascade menu for the "On Tk" item above.

menu .menu.help.m.tk
.menu.help.m.tk add command -label "Init File" \
                            -command {ShowFile $tk_library/tk.tcl}
.menu.help.m.tk add command -label "Library Location" \
        -command {SendToSoarShell {uplevel #0 {format "%s" $tk_library}}}
.menu.help.m.tk add command -label "Manual Pages" \
        -command {ShowManPage [uplevel #0 {format "%s" $tk_library}]/../doc}
.menu.help.m.tk add command -label "Version" \
               -command {SendToSoarShell {uplevel #0 {expr $tk_version}}}
.menu.help.m.tk add separator
.menu.help.m.tk add cascade -label "WWW" -menu .menu.help.m.tk.www

# Subcascade for "On Tk/WWW"

menu .menu.help.m.tk.www
.menu.help.m.tk.www add command -label "General" \
       -command "exec $wwwViewer http://www.isi.edu/soar/schwamb/tcl.html &"
.menu.help.m.tk.www add command -label "Manual Pages" \
  -command "exec $wwwViewer http://www.cis.upenn.edu/~ioi/tix/doc/tcltkman/tcltkman.html &"

# Cascade menu for the "On Tcl" item above.

menu .menu.help.m.tcl
.menu.help.m.tcl add command -label "Init File" \
                             -command {ShowFile [info library]/init.tcl}
.menu.help.m.tcl add command -label "Library Location" \
                             -command {SendToSoarShell "info library"}
.menu.help.m.tcl add command -label "Manual Pages" \
                     -command {ShowManPage [format {%s/../doc} [info library]]}
.menu.help.m.tcl add command -label "Version" \
                             -command {SendToSoarShell "info tclversion"}
.menu.help.m.tcl add separator
.menu.help.m.tcl add cascade -label "WWW" -menu .menu.help.m.tcl.www

# Subcascade for "On Tcl/WWW"

menu .menu.help.m.tcl.www
.menu.help.m.tcl.www add command -label "General" \
       -command "exec $wwwViewer http://www.isi.edu/soar/schwamb/tcl.html &"
.menu.help.m.tcl.www add command -label "Manual Pages" \
  -command "exec $wwwViewer http://www.cis.upenn.edu/~ioi/tix/doc/tcltkman/tcltkman.html &"

pack .menu.help -side right -padx 5

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

# Create MENU BAR

tk_menuBar .menu .menu.file .menu.show .menu.excise .menu.execute .menu.help

# Create TEXT WINDOW and scrollbar

text .t -relief raised -bd 2 -yscrollcommand ".s set" -setgrid true
scrollbar .s -relief flat -command ".t yview"

# QUICK ACTION BUTTONS

frame .frame -borderwidth 0
button .frame.run  -text "Run"  -command {schedule}
button .frame.step -text "Step" -command {schedule 1} 
button .frame.stop -text "Stop" -command {stop-soar}
button .frame.print -text "Goal Stack" -command {print -stack}

# Slider for simulation speed
scale .frame.simspeed -from 0.0 -to 10.0 -length 170 -orient horizontal -label "Simulation RootSpeed" -tickinterval 2.0 -showvalue true -width 10 -resolution 0.1 -bigincrement 1.0 -command {sim-speed}
bind .frame.simspeed <Any-Button> {focus %W}

# Make clicks move slider by bigIncrement (proc from scale.tcl)
#proc soarScaleButtonDown {w x y} {
#    global tkPriv
#    set tkPriv(dragging) 0
#    set el [$w identify $x $y]
#    if {$el == "trough1"} {
#	tkScaleIncrement $w up big initial
#    } elseif {$el == "trough2"} {
#	tkScaleIncrement $w down big initial
#    } elseif {$el == "slider"} {
#	set tkPriv(dragging) 1
#	set tkPriv(initValue) [$w get]
#	set coords [$w coords]
#	set tkPriv(deltaX) [expr $x - [lindex $coords 0]]
#	set tkPriv(deltaY) [expr $y - [lindex $coords 1]]
#    }
#}
#bind .frame.simspeed <Button-1> {
#    soarScaleButtonDown %W %x %y
#}

# Packup the GUI

#pack .frame.run .frame.step .frame.stop .frame.print -side left -padx 5 -pady 5
pack .frame.run .frame.step .frame.stop .frame.print .frame.simspeed -side top -fill x
#pack .frame -side bottom -fill x
pack .frame -side left -fill y
pack .s -side right -fill both
pack .t -side left -expand 1 -fill both

#--------------------------------------------------------------------------

# BINDINGS

# Set up bindings for text window 

# Control-C stops run
bind .t <Control-c> stop-soar

# Select a single character and set focus to text widget

bind .t <1> {
    set tk_priv(selectMode) char
    %W mark set anchor @%x,%y
    if {[lindex [%W config -state] 4] == "normal"} {focus %W}
}

# Delete the current line

bind .t <Control-u> {
    .t delete {promptEnd + 1 char} insert
    .t yview -pickplace insert
    set inputString ""
}

# Invoke a command if complete

bind .t <Return> {.t insert insert \n; AddToInput \n; Invoke; break}

# Remove a single character

bind .t <BackSpace> {Backspace ; break}

bind .t <Delete> {Backspace ; break}

bind .t <Control-h> {Backspace ; break}

# Insert the current selection

bind .t <Button-2> {
    set sel [selection get]
    AddToInput $sel
    if [string match *.0 [.t index insert]] {
	Invoke
    }
}

# Add the current character to the inputString buffer.

bind .t <Any-KeyPress> {
    if {"%A" != ""} {
	AddToInput %A
    }
}

#--------------------------------------------------------------------------

# Support PROCEDURES

# The following procedures are listed in alphabetical order.

# Simply add an input character to the inputString buffer

proc AddToInput {chars} {
    global inputString
    set inputString $inputString$chars
}

# The following procedure below handles backspaces, being careful not
# to backspace over the prompt.  Reset the inputString buffer to
# match the new contents of the text widget.

proc Backspace {} {
    global inputString
    if {[string compare [.t index promptEnd] [.t index {insert - 1 char}]]} {
	.t delete insert-1c insert
	.t yview -pickplace insert
    }
    if {$inputString != ""} {
	set inputString \
            [string range $inputString 0 \
                    [expr [string length $inputString] - 2]]
    }
}

# The procedure below executes a command (it takes everything after 
# the prompt) by sending it to the currently selected agent.  In 
# contrast to SendToSoarAgent, this procedure is used when the
# user has already typed the command into the text widget.  We set
# the "executing" flag to avoid executing a command while we are
# already processing one.
#
# If a command is incomplete, we skip the invocation.  If it is
# complete, we check to see if we must intercept the command and
# treat it specially since we are operating under a GUI.  Since
# the history commands are not available in the GUI, we provide
# the most commonly used one here, the !! command to repeat the
# last command.  Other commands we intercept include "help" because
# the normal help command uses a pager in stdout.  Here we invoke
# the help used from the menus, instead.  For other commands, we
# send them to the currently selected agent.  We are careful to
# trap errors and print the results to the text widget and the
# log file (if open).  Once the results are printed, we reset the
# inputString buffer to receive the next command and issue a
# prompt.
#
# If we are logging, we copy the command and its results to the
# the log file.  Since we are not using the normal command line
# interpreter, we must explicitly add these items to the log file.

proc Invoke {} {
    global app executing lastCommand inputString logging soar_library
    global documented_aliases
    set cmd $inputString
    incr executing 1

    if [info complete $cmd] {
	if {$cmd == "!!\n"} {
	    set cmd $lastCommand
	} else {
	    set lastCommand $cmd
	}
	if {$logging == 1} {send $app [list log -add $cmd]}

# If we see "help" followed by at least one space, we snag all
# subsequent characters until we see a space, carriage return,
# or tab (that explains the odd formating below).  Those characters
# should be the name of a help topic.

	if {([regexp -nocase {help[ ]+([^ 
	]+)} $cmd ignore topic])
	    && ($topic != "-all")} {
		if {[file isfile $soar_library/../doc/cat/$topic.n] == 1} {
		    ShowFile $soar_library/../doc/cat/$topic.n 0
		} else {
		    .t insert insert "Sorry, there is no help on \"$topic\".\n"
		}
        } elseif {[regexp -nocase {^[ ]*select-interp[ ]+([^ 
	    ]+)} $cmd ignore name]} {
	    SelectAgent $name
	} else {
	    if {[regexp -nocase {^[ ]*log[ ]+-open} $cmd ignore]} {
		set logging 1
	    }
	    if {[regexp -nocase {^[ ]*log[ ]+-close} $cmd ignore]} {
		set logging 0
	    }
	    set result [catch [list send $app $cmd] msg]
	    if {$result != 0} {
		.t insert insert "ErrorXX: $msg\n"
		if {$logging == 1} {
		    send $app [list log -add "Error: $msg\n"]
		}
	    } else {
		if {$msg != ""} {
		    .t insert insert $msg\n
		    if {$logging == 1} {
			send $app [list log -add "$msg\n"]
		    }
		}
	    }
	}
        set inputString ""

	Prompt
    }

    incr executing -1
    .t yview -pickplace insert
}

# This procedure is invoked when the user clicks on the logging
# checkbutton in the FILE menu.

proc LogAction {} {
    global logging app
    if {$logging == 0} {
	catch {send $app {log -close}}
    } else {
	set file [FSBox]
	if {$file != ""} {
	    catch {send $app [list log -open $file w]}
	} else {
	    after 0 {uplevel #0 {set logging 0}}
	}
    }
}

# This procedure simply throws up a notification dialog to let the
# user know that some capability is not implemented yet.

proc NotYet {} {
    tk_dialog .dlg "TkSoar Notification" \
              "Sorry, that capability is not implemented yet." \
              error 0 {Dismiss}              
}

# This procedure takes a command that contains a reference to 
# symbol "$file".  The user is queried for a file name and the
# given command is executed with the context set so that the
# symbol "$file" contains the selected file name.  If an error
# occurs while processing the given command, we notify the 
# user.

proc ProcessFile {method} {
    set file [FSBox]
    if {$file != ""} {
	set result [catch {eval $method} msg]
	if {$result != 0} {
	    .t insert insert "Error: $msg\n"
	}
	return $result
    }
    return 0
}

# The procedure below is used to print out a prompt at the insertion 
# point.  If the command entered by the user is incomplete, then no 
# prompt is printed.  If the insertion point is not at the beginning
# of a line, a carriage return is added so the prompt will always 
# appear at the left.  If we are logging, then we dump the prompt
# to the log file as well.  Finally, highlight the prompt by
# marking it as bold text and set the promptEnd mark so the cannot
# backspace over the prompt.  The font used by the bold tag is
# defined in the initialization section.

proc Prompt {} {
    global app inputString logging
    set cmd $inputString 
    if [info complete $cmd] {
	scan [.t index insert] "%d.%d" line column
	if {$column != 0} {.t insert insert "\n"}
	.t insert insert "$app> "
	if {$logging == 1} {
	    if {$column != 0} {
		send $app [list log -add "\n"]
	    }
	    send $app [list log -add "$app> "]
	}
	.t mark set promptEnd insert-1c
	.t tag add bold {promptEnd linestart} promptEnd
    }
}

# This procedure converts the text in a text widget from a raw nroff
# output form to a prettier form.  The text in the widget is scanned
# for nroff output format controls such as the following:
#
#   Underscored character: <underscore><backspace><char>
#   Bolded character: <char><backspace><char><backspace><char><backspace><char>
#   Trash: <escape>9
#
# Since this scanning process can be slow, we update idletasks as we
# process the lines.  The basic process is to find one of the above
# sequences, delete the control characters, and highlight the character
# according to its intended format by using text widget tags.
#
# This procedure needs to be made faster.  This could be done by 
# modifying a line for all changes before writting characters back
# into the text widget.  This requires that we generate a table mapping
# character indices to formatting tags and then using that to adjust
# the text widget.

proc ScanTextAndHighlight {w} {
    $w config -cursor watch
 
    scan [$w index end] %d numLines
    set begin 0

    for {set i 1} {$i < $numLines} {incr i} {
	set text [$w get $i.$begin "$i.$begin lineend"]  
        set check_line 1

        # this could take a while - so keep up to date on events!
        update 

        while {$check_line} {

	    set check_line 0

	    if {[regexp -indices -nocase "\x1b\[9\]" \
		    [string range $text $begin end] indices] == 1} {
                set left [expr $begin + [lindex $indices 0]]
                set right [expr $begin + [lindex $indices 1]]
                $w delete $i.$left $i.$right+1c
		
                set text [$w get $i.$begin "$i.$begin lineend"] 
	        set check_line 1
            }

	    if {[regexp -indices -nocase "(.)\x08(.)\x08(.)\x08(.)" \
		    [string range $text $begin end] indices one two three four] == 1} {
		set one   [string range $text [lindex $one 0] [lindex $one 1]]
		set two   [string range $text [lindex $two 0] [lindex $two 1]]
		set three [string range $text [lindex $three 0] [lindex $three 1]]
		set four  [string range $text [lindex $four 0] [lindex $four 1]]
		if {([string compare $one $two] == 0) &&   \
			([string compare $one $three] == 0) && \
			([string compare $one $four] == 0)} {
		    set left [expr $begin + [lindex $indices 0]]
		    set right [expr $begin + [lindex $indices 1]]
		    $w delete $i.$left $i.$right

		    $w tag add bold "$i.$left" "$i.$left+1c"

		    set text [$w get $i.$begin "$i.$begin lineend"] 
		    set check_line 1
		}
	    }

	    if {[regexp -indices -nocase "_\x08." \
		    [string range $text $begin end] indices] == 1} {
		set left [expr $begin + [lindex $indices 0]]
		set right [expr $begin + [lindex $indices 1]]
		$w delete $i.$left $i.$right
		
		$w tag add italic "$i.$left" "$i.$left+1c"

		set text [$w get $i.$begin "$i.$begin lineend"] 
		set check_line 1
	    }
	}

    }

     $w tag configure bold -font *-Courier-Bold-R-Normal-*-120-*
     $w tag configure italic -font *-Courier-Medium-O-Normal-*-120-*

     $w config -cursor xterm
}

# The following procedure is invoked to change the agent/application
# that we're talking to.  It also updates the prompt for the current
# command, unless we're in the middle of executing a command from
# the text item (in which case a new prompt is about to be output
# so there's no need to change the old one).

proc SelectAgent appName {
    global app executing
    set found 0
    foreach i [list-interps -agent] {
	if {$i == $appName} {
	    set app $appName
	    set found 1
	}
    }
    if !$found {
	.t insert insert "Agent $appName does not exist"
	return
    }
    if !$executing {
	.t delete "promptEnd linestart" promptEnd
	.t insert promptEnd "$appName>"
	.t tag add bold "promptEnd linestart" promptEnd
    }
    UpdateLogFlag
    return {}
}

# This procedure sends a command to the currently selected agent's
# interpreter.  This command is used from menus and simulates the 
# user typing in the same command.  Hence, the command and its 
# results are pasted into the text window and sent to the log file
# (if open).  The text window is also scrolled so the user can see
# the end of the result.  This approach helps new users learn the
# textual form of Soar commands and helps all users log actions
# whether typed in or invoked from menus.

proc SendToSoarShell {command} {
  global app logging
  .t insert insert "$command\n"
  if {$logging == 1} {
      send $app [list log -add "$command\n"]
  }
  set result [send $app "$command"]
  .t insert insert "$result"
  if {$logging == 1} {
      send $app [list log -add "$result\n"]
  }
  Prompt
  .t yview -pickplace insert
}

# The ShowFile procedure shows a file to the user by creating a
# new text window and placing the contents of the file in the 
# window.  If an additional argument is given, it indicates
# whether a nroff man_page is being shown or not.  If not, then 
# we just dump the file contents into the window.  If so, then
# we first run nroff on the file and remove backspaces and 
# escapes.

proc ShowFile {file {man_page 0}} {
    global fileWindowCounter
    if {[file isfile $file] == 1} {
	set w ".showfile$fileWindowCounter"
	incr fileWindowCounter

	catch {destroy $w}

	toplevel $w
	wm title $w  $file
	wm iconname $w $file
	wm minsize $w 0 0

	text $w.text -width 85 -setgrid 1 -bd 2 \
                     -yscrollcommand "$w.scroll set" 
	scrollbar $w.scroll -command "$w.text yview"
	button $w.ok -text Dismiss -command "destroy $w"

	pack $w.ok -side bottom -fill x
	pack $w.scroll -side right -fill y
	pack $w.text -side left -expand 1 -fill both

	$w.text delete 1.0 end
	if {$man_page == 0} {
	    set f [open $file]
	    while {![eof $f]} {
		$w.text insert end [read $f 1000]
	    }
	    close $f
	    ScanTextAndHighlight $w.text
	} else {
	    cd [file dirname $file]
	    regsub -all (\x08|(\x1b)9) [exec nroff -man $file] "" page
	    $w.text insert end $page
	}
    }
}

# This procedure asks the user for a man page to display and 
# then shows it to the user in a separate window.

proc ShowManPage {dir {use_nroff 1}} {
    set file [FSBox "Select file:" $dir]
    if {$file != ""} {
	ShowFile $file $use_nroff
    }
}

# This procedure updates the logging flag to match the logging
# state of an agent when we select a new agent.

proc UpdateLogFlag {} {
    global logging app
    if {[send $app {log -query}] == "open" && !$logging} {
	set logging 1
    }
    if {[send $app {log -query}] != "open" && $logging} {
	set logging 0
    }
}

#-------------------------------------------------------------------

# Startup INITIALIZATIONS:

# Redirect all printed output (from stdout, usually) to the text widget

output-strings-destination -push -text-widget .t

# We add a monitor to make sure that long running commands like
# "go" do not stop X events from being processed.

monitor -add after-decision-cycle update

# Define the text tag "bold" to indicate the following font.

.t tag configure bold -font *-Courier-Bold-R-Normal-*-120-*

# NOTE: to use a smaller font, try:
#.t config -font -*-Courier-Medium-R-Normal--*-90-*-*-*-*-*-*
# ...and set all the other fonts to 90 pts also.

# Tell the window manager the name of this GUI and its icon.

global interp_name
wm title . $interp_name
wm iconname . $interp_name

# If the window is hidden, make sure we open it and make it
# visible.

wm deiconify .

# Set the initial state of the text widget by viewing the top
# and printing a prompt.

.t mark set promptEnd insert-1c
.t yview -pickplace insert
Prompt

# Reset the focus so that keyboard events are directed at the
# text widget.

focus .t


