
/*************************
  command timing:  cmd_timing.c
  Behnam Salemi, modified by Gal Kaminka
 ************************/

#include <sys/time.h>
#include <sys/timeb.h>
#include "robocup.h"
#include "soar2c.h"
#include <soar.h>

#define abs(x) ( x < 0 ? -(x) : x )

#define   __TIMING_EXTERN__
#include  "cmd_timing.h"

/* #define ISSUE_COMMAND_DELAY    65  minimum # of ms delay between output commands */



static struct timeb old_time, current_time;
static Bool action_time = TRUE;

static Symbol *issue_command_sym, *issue_command_yes_sym,*issue_command_no_sym;
static wme * issue_command_wme;

static Symbol *copy_top_input_link_id;

void cmd_timing_init(Symbol *top_input_link_id)
{
  /* initialize timing wmes */
  
  issue_command_yes_sym   = get_io_sym_constant("*yes*");
  issue_command_no_sym    = get_io_sym_constant("*no*");
  copy_top_input_link_id  = top_input_link_id; 

  issue_command_wme = CREATE_WME_ATTRIB_CONST(copy_top_input_link_id,
					      issue_command_sym, 
					      get_io_sym_constant ("action-time"),
					      issue_command_yes_sym);

  action_time = TRUE;

  if (-1 == ftime(&old_time))
    {
      fprintf(stderr,"cmd_timing.c/cmd_timing_init: Can't get system time\n");
      exit(1);
    }

  return ;
}
    
void start_action_timer()
{

  /* This function sets the action-time flag to no, 
     preventing more output at this time */
  /*
  UPDATE_WME_CONST(issue_command_wme,
		   copy_top_input_link_id, issue_command_sym,
		   issue_command_no_sym);
		   */
		   
  action_time = FALSE;
  if (-1 == ftime(&old_time))
    {
      fprintf(stderr,"cmd_timing.c/start_action_timer: Can't get system time\n");
      exit(1);
    }
#ifdef DEBUG
  fprintf(stderr,"setting action_time to *no*\n");
#endif
}

void check_action_timer()
{
  /* if the current_time is greater than old_time by Issue_command_delay ms,
     enable the action-time flag */

  int ms_diff;

  if (FALSE == action_time)
    {
      if (-1 == ftime(&current_time))
	{
	  fprintf(stderr,"cmd_timing.c/check_action_timer:  Can't get system time\n");
	  exit(1);
	}

      ms_diff = (current_time.time*1000 + current_time.millitm)
	- (old_time.time*1000 + old_time.millitm);

      if (ms_diff > Issue_command_delay)
	{
	  /*
	  UPDATE_WME_CONST(issue_command_wme,
			   copy_top_input_link_id, issue_command_sym,
			   issue_command_yes_sym);
			   */
	  action_time = TRUE;
#ifdef DEBUG	  
	  fprintf(stderr, "setting action_time to: *yes*\n");
#endif      
	}
    }
  else
    {
#ifdef DEBUG
      fprintf(stderr, "can't set action_time.  Too early\n");
#endif
    }

  return ;
}


void wait_until_action_time()
{
  /* unless the current_time is greater than old_time by Issue_command_delay ms, wait */
  static struct timeb my_old_time;
  int ms_diff;

  if (FALSE == action_time)
    {
      if (-1 == ftime(&current_time))
	{
	  fprintf(stderr,"cmd_timing.c/check_action_timer:  Can't get system time\n");
	  exit(1);
	}
      
	ms_diff = (current_time.time*1000 + current_time.millitm)
	  - (old_time.time*1000 + old_time.millitm);
	
      if (ms_diff < Issue_command_delay)
	{
	  /* then wait DELAY */
	  usleep(Issue_command_delay*1000);

	  /*
	  if (-1 == ftime(&my_old_time))
	    {
	      fprintf(stderr,"cmd_timing.c/start_action_timer: Can't get system time\n");
	      exit(1);
	    }
	  
	  do {
	    if (-1 == ftime(&current_time))
	      {
		fprintf(stderr,"cmd_timing.c/check_action_timer:  Can't get system time\n");
		exit(1);
	      }

	    ms_diff = (current_time.time*1000 + current_time.millitm)
	      - (my_old_time.time*1000 + my_old_time.millitm);

	  } while (ms_diff < Issue_command_delay);
	  */
	}
	/*
	  if (ms_diff >= Issue_command_delay)
	  {
	  ;
	  }
	  else
	  {
	  */
      /* fprintf(stderr,"sleeping %d ms\n", ms_diff); */
      /* usleep(ms_diff); */
      /* usleep((Issue_command_delay-ms_diff)*950); /* sleep 95% of remaining msec */
      /* usleep(Issue_command_delay*1000); */
      /*
           }
	   */
      action_time = TRUE;
    }
  
  
  return ;
}



