

/********************************************************************
 * Intercept Utilities
 ********************************************************************/

#include <math.h>

#define __INTERCEPT_EXTERN__
#include "robocup.h"  /* for role macro definition */
#include "intercept.h"



void intercept_init(void)
{
  Incrddir          = INCRDDIR;
  Dashinc           = DASHINC;
  Dashinc2           = DASHINC2;
  Dashoff           = DASHOFF;
  Maxdash           = MAXDASH;
  Close             = CLOSE;
  Almost            = ALMOST;
  Almost_straight   = ALMOST_STRAIGHT;
  Thug_dx           = THUG_DX;
  return ;
}


static void go_next (double dx, double dir, double ddx, double ddir, 
		     double *turn, double *dash)
{
    double wturn, wdash, incdir;

    incdir = (Incrddir * ddir);
    incdir = (incdir > ANG45) ? ANG45 : incdir;
    incdir = (incdir < -ANG45) ? -ANG45 : incdir;
    wturn = dir + incdir;

    wdash = ((dx + ddx) * Dashinc) + Dashoff;
    wdash = (wdash > Maxdash ) ? Maxdash : wdash;
    
    *turn = wturn;
    *dash = wdash;

    return ;
}

/*

static void go_hard (double dx, double dir, double ddx, double ddir, 
		     double *turn, double *dash)
{
    double wturn, wdash, incdir;
    incdir = (Incrddir * ddir);
    incdir = (incdir > ANG45) ? ANG45 : incdir;
    incdir = (incdir < -ANG45) ? -ANG45 : incdir;
    wturn = dir + incdir;
    wdash = (dx * Dashinc) + Dashoff;
    wdash = (wdash > Maxdash ) ? Maxdash : wdash;

    *turn = wturn;
    *dash = wdash;

    return ;
}

static void go_slow (double dx, double dir, double ddx, double ddir, 
		     double *turn, double *dash)
{
    double wturn, wdash, incdir;

    incdir = (Incrddir * ddir);
    incdir = (incdir > ANG45) ? ANG45 : incdir;
    incdir = (incdir < -ANG45) ? -ANG45 : incdir;
    wturn = dir + incdir;

    wdash = Dashoff;
    wdash = (wdash > Maxdash ) ? Maxdash : wdash;

    *turn = wturn;
    *dash = wdash;

    return ;
}

*/

static void go_shortest(double dx, double dir, double ddx, double ddir, 
			double *turn, double *dash)
{
    double ballvel, disttogo, sinq, sinang, cosang, me_togo, ball_togo, cosq, q, likelydir, wturn, wdash, rel_dx;
    double odx = dx - ddx;
    /* We are reasoning in a geometric space so let's avoid negative angles */
    double posddir = abs(ddir) ;
    double ang = posddir*PI/180.0; 
    int  i;

    /* ang is the angle in radians that the ball travels
       through from the perspective of the player
       */
    cosang =  cos(ang) ;
    sinang =  sin(ang) ;

    /* Calc the Ball Vel */
    /* should be standard use of cosine law (and should be sqrt of a pos num!*/

    ballvel = sqrt( (dx * dx) + (odx * odx) - ( 2 * dx * odx * cosang ) );


      /* it looks like q is supposed to be the angle between the ball's dir
	 of travel and the dir in which the player most recently sighted it  */

      sinq = ( odx * sinang) / ballvel ;
      q = asin(sinq);
      
      /* now we have to just complete the right triangle, q is supplemental
	 to one of the angles and the dir to turn so as to travel along the
	 perpendicular is wturn. So.... 
	 The call to abs is simply precautionary - hopefully it is low cost */

      wturn =  (PI/2) - q ;
      ball_togo = abs(dx * sin(wturn)) ;
      me_togo = abs(dx * cos(wturn)) ;
      
      /* Now we can decide on how to turn */
      /* IF Close then head directly */
      if (me_togo < Close || ball_togo < Thug_dx) { 
	wturn = dir + (Incrddir*ddir);
      } else {
	/* Convert back to degrees and restore negative angle reasoning */
	  /* And average with where the ball is likely to be */
	wturn = (wturn * (180/PI)) ;
	if (ddir > 0) {
	    if 	( (wturn > (4.0 * ddir)) || (wturn < (2.0 * ddir) ) ) {
		wturn = dir + ((wturn + (ddir * 3)) / 2.0 );
	    } else {
		wturn = wturn + dir;
	    }
	} else {
	    if 	( (wturn < (4.0 * ddir)) || (wturn > (2.0 * ddir) ) ) {
		wturn = dir + (((ddir * 3) - wturn) / 2.0 ) ;
	    } else {
		wturn = wturn + dir;
	    }
	}
      }
      /* to determine the DASH power, lets put together all we know
	 about how far player and ball must travel and Ball velocity.
	 Note we don't know player velocity */

      /*      rel_dx = (me_togo * ballvel) / ball_togo ;
	      wdash = (rel_dx * Dashinc2) + Dashoff; */

      wdash = (abs(me_togo) * Dashinc2) + Dashoff;

      wdash = (wdash > Maxdash ) ? Maxdash : wdash;

      *turn = wturn;
      *dash = wdash;
      
      return ;

}


/* GalK:  Should really also return likelyhood! XXXXXXX */
/* Stacy:  Okay, the likelihood is 0. */
void intercept(double dx, double dir, double ddx, double ddir,
	       double absx, double absy,
	       double absdir, int player, double *turn, double *dash)
{

    /* this is here just to catch weird bugs from the Soccer Server 
       (may only be necessary in coach mode */
    *turn = 0;
    *dash = 0;

    if (  (dx < -359.0) || (dir < -359.0) ) return;
    *turn = dir;
    *dash = 100;
    if ( (ddx > 1023.0) || (ddir > 1023.0) ) return;
    


  if (dx <= Almost ||
      (-Almost_straight <= ddir && ddir <= Almost_straight) ||
      ddx > 0) {
      go_next(dx,dir,ddx,ddir,turn,dash); 
  } else {
      go_shortest(dx,dir,ddx,ddir,turn,dash);
  }


  if (player == GOALEE && (Goal_Loc - .5 < absx) && (Goal_Loc + 2 > absx) ) {
      absdir = *turn + absdir;
      if (absdir > 90 && absdir < 180) {
	  *turn = 90;
	  *dash = *dash * (90 / absdir); /* OKAY - DONT LAUGH */
      } else if (absdir < -90 && absdir > -180) {
	  *turn = -90;
	  *dash = *dash * (-90 / absdir); /* OKAY - DONT LAUGH */
      }
  }	  


  return ;
  
}




