/* File   : parser.c
   Info   : parses soccer server messages and updates data structures
   Author : Ali Erdem
   Date   : April 30, 1997
*/

#include "robocup.h"
#include <math.h>


#include <sys/time.h>
#include <sys/timeb.h>

static void print_state(int);


void parse (char *line) 
{
   static SensorInfo parsed_line;
   register int i, j;
   int player_no, team_no, fpos, lpos, team;
   extern double opponent_goal_center(void); /* kick_dir.c */
   extern void regions_init(void); /* clear_ball.c */
   extern void update_team_player(double, double); /* clear_ball.c */
   extern void update_opponent_player(double, double); /* clear_ball.c */
   extern double clear_ball_angle(void); /* clear_ball.c */
   extern Bool location(double, double, double, double, double,
			double *, int *);  /* location.c */

   int ms_diff;

   static Bool first_time = TRUE;
   static struct timeb old_time, current_time;
   
#ifdef DEBUG
   printf("parser.c/parse: Received %s\n",line);
#endif

   if (first_time)
     {
       if (  (-1 == ftime(&old_time))
	   ||(-1 == ftime(&current_time)))
	 {
	   fprintf(stderr,"parser.c can't get system time.\n");
	   exit(1);
	 }
       first_time = FALSE;
     }

   if (scan_info(line,&parsed_line,Teamname)) {
     if (parsed_line.type == SIT_See)
       {
	 /*
	 if (-1 == ftime(&current_time))
	   {
	     fprintf(stderr,"parser.c:  Can't get system time\n");
	     exit(1);
	   }
	 
	 ms_diff = (current_time.time*1000 + current_time.millitm)
	          -(old_time.time*1000+old_time.millitm);
	 
	 fprintf(stderr,"parser.c: time between sense cycles: %d ms\n",ms_diff);
	 
	 old_time = current_time;
	 */

	 /* estimate current position from observed objects */
	 if (!estimate_current_pos(&parsed_line.info.see,MySide,&(State[NOW].Players[OWN][MyNumber-1].position)))
	   {
	     /* failed! */
	     /* State[NOW].Players[OWN][MyNumber-1].position.dir = DIR_ERR; */

	     /* XXX DEBUGGING TEST
	     fprintf(stderr,"----> Player %d Theta=%4.2f X=%4.2f Y=%4.2f", MyNumber, 
		     State[NOW].Players[OWN][MyNumber-1].position.dir,
		     State[NOW].Players[OWN][MyNumber-1].position.pos.x,
		     State[NOW].Players[OWN][MyNumber-1].position.pos.y);
		     */
	     /* instead, use old one */
	     State[NOW].Players[OWN][MyNumber-1].position.pos.x = State[PREVIOUS].Players[OWN][MyNumber-1].position.pos.x;
	     State[NOW].Players[OWN][MyNumber-1].position.pos.y = State[PREVIOUS].Players[OWN][MyNumber-1].position.pos.y;
	     State[NOW].Players[OWN][MyNumber-1].position.dir   = State[PREVIOUS].Players[OWN][MyNumber-1].position.dir;

#ifdef DEBUG
	     fprintf(stderr,"parser.c: Failed to calculate own position\n");
#endif
	   }
#ifdef DEBUG
	 fprintf(stderr,"parser.c: %lf %lf %lf\n",State[NOW].Players[OWN][MyNumber-1].position.pos.x,
		 State[NOW].Players[OWN][MyNumber-1].position.pos.y,
		 State[NOW].Players[OWN][MyNumber-1].position.dir);
#endif

	 /* XXX DEBUGGING TEST
	 if (( State[NOW].Players[OWN][MyNumber-1].position.dir  < -181) ||
	     ( State[NOW].Players[OWN][MyNumber-1].position.dir > 181)) 
	   {
	     fprintf(stderr,"----> Player %d Theta=%4.2f X=%4.2f Y=%4.2f", MyNumber, 
		     State[NOW].Players[OWN][MyNumber-1].position.dir,
		     State[NOW].Players[OWN][MyNumber-1].position.pos.x,
		     State[NOW].Players[OWN][MyNumber-1].position.pos.y);
	     
	   }
	   */

	 /* update time from server */
	 State[NOW].time = parsed_line.info.see.time;
	 
	 /* reset visible player count needed by c4.5 kicking */
	 State[NOW].Auxilary.cPlayerCnt = 0;

	 /* reset all objects to not visible */
	 for (i=0; i<TEAMS; i++)
	   for (j=0; j<MAX_PLAYERS; j++)
	     State[NOW].Players[i][j].visible = FALSE;

	 State[NOW].Ball.visible = FALSE;
	 for (i=0; i<TEAMS; i++)
	   State[NOW].Goals[i].visible = FALSE;
	 for (i=0; i<NUM_FLAGS; i++)
	   State[NOW].Flags[i].visible = FALSE;
	 for (i=0; i<NUM_LINES; i++)
	   State[NOW].Lines[i].visible = FALSE;

#ifdef CLEAR_KICK
	 /* reset kick_dir calc */
	 regions_init();
#endif
	 
	 /* now update changes */
	 for (i=0;i<parsed_line.info.see.n_obj;i++) {
	   switch (parsed_line.info.see.obj[i].type) {
	   case OT_Ball:
	     State[NOW].Ball.visible = TRUE;
	     State[NOW].Ball.SeenInfo = parsed_line.info.see.obj[i];
	     /*
	     if (!estimate_object_pos(&(State[NOW].Ball))) {
	       State[NOW].Ball.position.dir = DIR_ERR;
#ifdef DEBUG
	       fprintf(stderr,"parser.c: Failed to calculate Ball position\n");
#endif
	     }
	     */
	     break;
	     
	   case OT_Player:
	     player_no = parsed_line.info.see.obj[i].id.player.unum;

	     /* keep track of seeing a player for calc. c4.5 kicking */
	     State[NOW].Auxilary.AnglesOfVisiblePlayers[State[NOW].Auxilary.cPlayerCnt] = 
	       parsed_line.info.see.obj[i].dir;
	     State[NOW].Auxilary.cPlayerCnt++;

#ifdef CLEAR_KICK	     
	     /* keep track of seeing a player:  kick_dir */

	     if (WT_Our == parsed_line.info.see.obj[i].id.player.team)
	       {
		 update_team_player(parsed_line.info.see.obj[i].dir,
				    parsed_line.info.see.obj[i].dist);
	       }
	     else 
	       {
		 update_opponent_player(parsed_line.info.see.obj[i].dir,
					parsed_line.info.see.obj[i].dist);
	       }
#endif

#ifdef DEUBG
	     fprintf(stderr,"received info about %d player, number %d\n",
		     parsed_line.info.see.obj[i].id.player.team, 
		     parsed_line.info.see.obj[i].id.player.unum);
#endif
	     if (WT_Unknown != parsed_line.info.see.obj[i].id.player.team)
	       {
	         if (WT_Our == parsed_line.info.see.obj[i].id.player.team) 
		   {
		     team = OWN;
		   }
		 else /* == WT_Opponent */
		   {
		     team = OPPONENT;
		   }
		 
		 if (player_no>0 && player_no<12)
		   {
		     State[NOW].Players[team][player_no-1].visible = TRUE;
		     State[NOW].Players[team][player_no-1].SeenInfo = parsed_line.info.see.obj[i];

		     /*
		       if (!estimate_object_pos(&(State[NOW].Players[team][player_no-1])))
		       fprintf(stderr,"parser.c: Failed to calculate player[%d][%d] position\n",team,player_no-1);
		       */
		   }
	       }
	   
	     break;
	     
	   case OT_Goal:
	     {
	       static int side, goal;
	       
	       side = parsed_line.info.see.obj[i].id.side;
	       if (side!=S_UNKNOWN)
		 {
		   if (MySide == side)
		     goal = OWN;
		   else 
		     goal = OPPONENT;
		   
		   State[NOW].Goals[goal].visible  = TRUE;
		   State[NOW].Goals[goal].SeenInfo = parsed_line.info.see.obj[i];
		 }
	     }
	     break;
	     
	   case OT_Flag:
	     fpos = (int)parsed_line.info.see.obj[i].id.fpos;
	     State[NOW].Flags[fpos].visible  = TRUE;
	     State[NOW].Flags[fpos].SeenInfo = parsed_line.info.see.obj[i];
	     break;
	     
	   case OT_Line:
	     lpos = (int)parsed_line.info.see.obj[i].id.lpos;
	     State[NOW].Lines[lpos].visible  = TRUE;
	     State[NOW].Lines[lpos].SeenInfo = parsed_line.info.see.obj[i];
	     break;
	     
	   case OT_Unknown:
	     fprintf(stderr,"parser.c: Unknown Object Type\n");
	     break;
	     
	   default:
	     fprintf(stderr,"parser.c: Error in Object Type\n");
	   }
	 }
	 
	 /* update best direction to kick */
	 /* XXX must do it AFTER estimate current_position is called */
	 State[NOW].Auxilary.opponent_goal_dir = opponent_goal_center();
	 State[NOW].Auxilary.clear_ball_angle  = clear_ball_angle();
       }
     else if (parsed_line.type == SIT_Hear)
       {
	 /* new time is given */
	 /* State[NOW].time = parsed_line.info.hear.time; */
	 memcpy(&(State[NOW].Msg), &(parsed_line.info.hear), sizeof(HearInfo));	 
	 switch (State[NOW].Msg.sender) {
	 case ST_Referee:
	   /* in case of goal announcements, filter out the point */
	   
	   /* following is a bit ugly, but I'll explain:
	       we need to detect "goal_<side>_<point>" msgs from the
	       referee, where <side> is "l" or "r", and <point> is the
	       point scored: 1,2,.... To do this we have to compare
	       the first word of the msg to "goal_l_" or "goal_r_" 
	       whose lengths are 6.

	       When we find such message, we cut off the _<point> suffix.
	       which is at location 6 = (strlen("goal_r"), for instance)
	     */
	   if (  (!strncmp(State[NOW].Msg.message,"goal_l", 6))
 	       ||(!strncmp(State[NOW].Msg.message,"goal_r", 6)))
	     State[NOW].Msg.message[6] = '\0'; /* cut message short */

	   fprintf(stderr,"Got referee msg: %s\n",State[NOW].Msg.message);
	   break;
	   
	 case ST_Self:     /* right now, we pass our own msgs and deal with them in the rules
			      interesting case of Corollary Discharge */
	   break;
	   
	 case ST_Player:
	   if (strncmp(parsed_line.info.hear.message,Teamname,strlen(Teamname)))
	     {
	       /* fprintf(stderr,"got peters msg and deleted\n"); */
	       State[NOW].Msg.message[0] = '\0';

	     }
	   break;
	   
	 }
       } else {
	 /* invalid message type */
	 fprintf(stderr,"parser.c: Invalid Message Type\n");
       }
   } else {
     /* parse failed */
     fprintf(stderr,"parser.c: Parse failed\nLine=%s\n",line);
   }
   
#ifdef DEBUG
   print_state(NOW);
#endif
}



void print_state (int step) {

     static int m=0;
     register int i,j;
     SeenObj sTmp;
     PosState pTmp;
/*
     if (m>20) {
        m = 0;
     } else {
       m++;
       return;
     }
*/
     printf("STATE AS OF CYCLE %d for player=%d\n",State[step].time,MyNumber);
     printf("================================================\n");
     for(i=0;i<2;i++) {
        for(j=0;j<11;j++) {
           sTmp = State[step].Players[i][j].SeenInfo;
           pTmp = State[step].Players[i][j].position;
           printf("PLAYER  Team:%1d  Player:%2d  team=%1d  UNum=%2d  InView=%1d  Dist=%5.2lf  Dir=%5.2lf  DistChng=%5.2lf  DirChng=%5.2lf  x=%5.2lf y=%5.2lf dir=%5.2lf\n",i,j,sTmp.id.player.team,sTmp.id.player.unum,sTmp.inViewConeP,sTmp.dist,sTmp.dir,sTmp.dist_chng,sTmp.dir_chng,pTmp.pos.x,pTmp.pos.y,pTmp.dir);
        }
     } 

     sTmp = State[step].Ball.SeenInfo;
     pTmp = State[step].Ball.position;
     printf("BALL  InView=%1d  Dist=%5.2lf  Dir=%5.2lf  DistChng=%5.2lf  DirChng=%5.2lf  x=%5.2lf y=%5.2lf dir=%5.2lf\n",sTmp.inViewConeP,sTmp.dist,sTmp.dir,sTmp.dist_chng,sTmp.dir_chng,pTmp.pos.x,pTmp.pos.y,pTmp.dir);

     for(i=0;i<TEAMS;i++) {
        sTmp = State[step].Goals[i].SeenInfo;
        printf("GOAL:%1d  Side=%1d  InView=%1d  Dist=%5.2lf  Dir=%5.2lf  DistChng=%5.2lf  DirChng=%5.2lf  x=%5.2lf y=%5.2lf dir=%5.2lf\n",i,sTmp.id.side,sTmp.inViewConeP,sTmp.dist,sTmp.dir,sTmp.dist_chng,sTmp.dir_chng,pTmp.pos.x,pTmp.pos.y,pTmp.dir);
     }

     for(i=0;i<NUM_FLAGS;i++) {
        sTmp = State[step].Flags[i].SeenInfo;
        printf("FLAG:%1d  fpos=%2d  InView=%1d  Dist=%5.2lf  Dir=%5.2lf  DistChng=%5.2lf  DirChng=%5.2lf  x=%5.2lf y=%5.2lf dir=%5.2lf\n",i,sTmp.id.fpos,sTmp.inViewConeP,sTmp.dist,sTmp.dir,sTmp.dist_chng,sTmp.dir_chng,pTmp.pos.x,pTmp.pos.y,pTmp.dir);
     }

     for(i=0;i<NUM_LINES;i++) {
        sTmp = State[step].Lines[i].SeenInfo;
        printf("LINE:%1d  lpos=%2d  InView=%1d  Dist=%5.2lf  Dir=%5.2lf  DistChng=%5.2lf  DirChng=%5.2lf  x=%5.2lf y=%5.2lf dir=%5.2lf\n",i,sTmp.id.lpos,sTmp.inViewConeP,sTmp.dist,sTmp.dir,sTmp.dist_chng,sTmp.dir_chng,pTmp.pos.x,pTmp.pos.y,pTmp.dir);
     }

     printf("HEARD  time=%5d  sender=%1d  dir=%5.2lf  msg=%s\n\n",State[step].Msg.time,State[step].Msg.sender,State[step].Msg.direction,State[step].Msg.message);

}







