/* File   : robocup.h
   Info   : Soccer simulation competition header file
            combines Ernesto.h, Gal Kaminka and Ali Erdem's changes
   Authors: Ernesto, Gal Kaminka, Ali Erdem
   Date   : April 25, 1997
*/

#include <stdio.h>
#include <libsclient.h>

#ifndef ROBOCUP
#define ROBOCUP

/* GalK:  Following is a trick to get only robocup.c to use the global vars as non-extern */
#ifndef __ROBOCUP_EXTERN__
#define __ROBOCUP_EXTERN__ extern
#endif

/* Global Variables */

__ROBOCUP_EXTERN__ Socket gSock;      /* connection to server */
__ROBOCUP_EXTERN__ char Server[120];  /* server name */
__ROBOCUP_EXTERN__ int Port;          /* port number on server */
__ROBOCUP_EXTERN__ int MySide;        /* S_LEFT, S_RIGHT, S_UNKNOWN */
__ROBOCUP_EXTERN__ int MyNumber;      /* My shirt number */
__ROBOCUP_EXTERN__ char Teamname[50]; /* Team name to use */

__ROBOCUP_EXTERN__ int GlobalRole;    /* role for global use, can be one of the following */
#define DEFENDER   0
#define ATTACKER   1
#define GOALEE     2


#define CLEAR_KICK

/* following for array indexing in the state */

#define TEAMS         2   /* own, and opponent */
#define OWN           0
#define OPPONENT      1

#define MAX_PLAYERS  11
/* following is for C4.5 kicking angle */
#define  _MAX_VISIBLE_PLAYERS_    ((TEAMS*MAX_PLAYERS)-1) /* can't see self */

/* players have seen info and calculated positions */

typedef struct _ObjInfo {
  SeenObj  SeenInfo;
  PosState position;
  Bool visible; /* currently visible or not */
} ObjInfo;


/* interception information */
typedef struct {
  double angle;
  double power;
  double likelihood;
} InterceptInfo;

/* auxilary calculated information */
typedef struct {
  InterceptInfo Intercept;           /* Interception possibility */
  double        opponent_goal_dir;   /* angle to opponent's goal */
  double        clear_ball_angle;    /* angle to clear the ball */
  int           num_friendlies_closer; /* # of teammates closer to ball */
  int           num_enemies_closer;    /* # of opponents closer to ball */
  double        home_direction;
  int           home_power;
  float         AnglesOfVisiblePlayers[_MAX_VISIBLE_PLAYERS_]; /* c4.5 kicking */
  int           cPlayerCnt; /* c4.5 kicking */
  float         smart_kick_angle; /* c4.5 current kick angle */
}  AuxInfo;


#define NUM_FLAGS    ((int)FP_P_R_B+1)              /* 13 flags enumerated in _FlagPos */
#define NUM_LINES    ((int)LP_R+1)                  /*  5 number of lines */

typedef struct {
  int      time;                                /* Simulation time */
  ObjInfo  Players[TEAMS][MAX_PLAYERS];         /* 2 teams */
  ObjInfo  Ball;                                /* 1 ball  (hopefully! ;-)) */
  ObjInfo  Goals[TEAMS];                        /* 2 goals */
  ObjInfo  Flags[NUM_FLAGS];                    /* 13 flags enumerated in _FlagPos */
  ObjInfo  Lines[NUM_LINES];                    /* 5 lines enumerated in _LinePos */
  HearInfo Msg;                                 /* Can hear a message in each cycle */
  AuxInfo  Auxilary;                            /* Auxilary Information */
} state;


#define NUM_OF_STATES  2
#define NOW      1
#define PREVIOUS 0      /* don't change this.  Should always be NOW-1 */

__ROBOCUP_EXTERN__ state State[NUM_OF_STATES];
__ROBOCUP_EXTERN__ PosState HomePosition;  /* coordinates of home position */


/* Macros & functions to map between the new structure and Ernesto's structure */

#define visible(x)        (x.visible == TRUE)

#define has_data_changed(x,y)  (x.dist != y.dist || x.dir != y.dir || x.dist_chng != y.dist_chng || x.dir_chng != y.dir_chng)


#endif


