/* 
 * soarAppInit.c --
 *
 *	Provides a default version of the main program and Soar_InterpInit
 *	procedure for Soar applications.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 * Copyright (c) 1995 Carnegie Mellon University, University of Michigan,
 *                    University of Southern California/Information Sciences
 *                    Institute.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * The Soar consortium proclaims this software is in the public domain, and
 * is made available AS IS.  Carnegie Mellon University, The University of 
 * Michigan, and The University of Southern California/Information Sciences 
 * Institute make no warranties about the software or its performance, implied
 * or otherwise.
 */

#include "robocup.h"
#include <soar.h>
#include <sys/time.h>


#define MAXMESG 512

#define MINTURN 10

#define MAXTURNDX 10

#define ACCEPT_INTERVAL 12000



/*
 * The following variable is a special hack that is needed in order for
 * Sun shared libraries to be used for Tcl.
 */

extern int matherr();
int *tclDummyMathPtr = (int *) matherr;


extern void my_input_function (soar_callback_agent a, 
			       soar_callback_data data, 
			       soar_call_data call_data);
extern void my_output_function (soar_callback_agent a, 
				soar_callback_data data, 
				soar_call_data call_data);

extern long Issue_command_delay;

/*
 *----------------------------------------------------------------------
 *
 * main --
 *
 *	This is the main program for the application.  We declare main
 *      here for two reasons.  First, the libraries created from the
 *      rest of the Soar code shouldn't contain "main" so that user
 *      code with existing main procedures can be easily linked against
 *      the library.  Second, placing main here allows the user to add
 *      additional code to the program for initialization and/or 
 *      command line processing before the primary Soar main procedure
 *      is invoked.
 *
 * Results:
 *	None: Soar_Main never returns here, so this procedure never
 *	returns either.
 *
 * Side effects:
 *	Whatever the application does.
 *
 *----------------------------------------------------------------------
 */

int
main(argc, argv)
    int argc;			/* Number of command-line arguments. */
    char *argv[];		/* Values of command-line arguments. */
{
  /* process RoboCup-specific arguments */
  
  /* Server's host name, RoboCup teamname to use */
  /*  extern char *Server, *Teamname; */  /* global lengh-limited arrays in ernesto.h */
  /* Server's port number to use */
  /* extern int Port; */

  int i;
  long tmplong;
  char * envptr;
  /* set defaults */
  strcpy(Teamname,"ISIS");
  strcpy(Server,"localhost");
  GlobalRole = DEFENDER;
  Port = 6000;
  /* Set the command delay from an environmental var */
  Issue_command_delay = 65;
  envptr = getenv("ISSUE_COMMAND_DELAY");
  if (envptr != NULL) {
    tmplong = strltol(envptr, NULL, 10);
    if (tmplong >= 0 && tmplong <= 200) {
        Issue_command_delay = tmplong;
    }
  }

  /*  for (i=0; i<argc; i++)
      printf("argv[%d] = %s\n",i, argv[i]); */
      
  i=0;
  while (i<argc)
    {
      int update = 0;

      if (!strcmp(argv[i], "-robocup-server"))
	{
	  strcpy(Server, argv[i+1]);
	  update = 1;
	}
      else if (!strcmp(argv[i], "-robocup-port"))
	{
	  Port = atoi(argv[i+1]);
	  update = 1;
	}
      else if (!strcmp(argv[i], "-robocup-teamname"))
	{
	  strcpy(Teamname, argv[i+1]);
	  update = 1;
	}
     else if (!strcmp(argv[i], "-robocup-role"))
	{
	  if (!strcmp(argv[i+1], "attack"))
	    GlobalRole = ATTACKER;
	  else if (!strcmp(argv[i+1], "defend"))
	    GlobalRole = DEFENDER;
	  else if (!strcmp(argv[i+1], "goalee"))
	    GlobalRole = GOALEE;
	  
	  update = 1;
	}


      if (1 == update)
	{
	  /* we have used an argument.  Get rid of it in argv so it doesn't
	     appear to be a Tcl/Tk/Soar argument later on */
	  int j;

	  argc -= 2; /* get rid of flag and content */
	  for (j=i; j<argc; j++)
	    {
	      argv[j] = argv[j+2];
	    }
	}
      else 
	i++;

    }

  printf("hostname is: %s  port: %d  teamname: %s \n",Server,Port,Teamname);

  /*  for (i=0; i<argc; i++)
      printf("argv[%d] = %s\n",i, argv[i]); */

  Soar_Main(argc, argv);
  return 0;			/* Needed only to prevent compiler warning. */
}

/*
 *----------------------------------------------------------------------
 *
 * Soar_InterpInit --
 *
 *	This procedure performs interpreter-specific initialization.
 *	Most applications, especially those that incorporate additional
 *	packages, will have their own version of this procedure.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	Depends on the added code.
 *
 *----------------------------------------------------------------------
 */

void 
Soar_InterpInit (interp, interp_type, a)
  Tcl_Interp * interp;		/* Interpreter to initialize.        */
  int          interp_type;     /* tclsh, wish, or agent interpreter */
  agent      * a;               /* Pointer to corresponding agent    */
{
  switch (interp_type) {
  case SOAR_ARGV_TCLSH:         /* Tcl (only) interpreter */
    break;
  case SOAR_ARGV_WISH:          /* Tcl interpreter with Tk extension */
    break;
  case SOAR_ARGV_AGENT:         /* Tcl interpreter with Soar extension */
                                /* (and optionally) Tk extension       */

    /* These two calls alter the low-level Soar I/O processing. */
    
    /* The change in input can be verified by printing portions */
    /* of working memory.  The top-state will have an attribute */
    /* called "io".  The value of the "io" attribute will be an */
    /* object with an attribute of "input-link".  The value of  */
    /* the input-link attribute will be an object which has an  */
    /* attribute called "number" whose value is another object. */
    /* This object will have the attribute "value" which will   */
    /* have an input number.  The structure is something like   */
    /* this (the identifier numbers may vary):                  */
    /*                                                          */
    /* S1 ^io I1                                                */
    /*    I1 ^input-link L1                                     */
    /*       L1 ^number O1                                      */
    /*          O1 ^value 3                                     */
    /*                                                          */
    /* Each time the agent is run and input occurs, the input   */
    /* function is called and the number is incremented.        */
    /* You can also verify that the symbols are cleaned up      */
    /* properly by performing an "init-soar" command and noting */
    /* that working memory is empty.  You can also verify that  */
    /* no symbol table leaks have occured due to input          */
    /* processing by executing the command "internal-symbols"   */
    /* which prints the contents of the symbol table.           */
    
    add_input_function(a, my_input_function, NULL, NULL, "test-input");
    
    /* The output function is run when changes occur on the     */
    /* Soar output link.  Output changes are made by adding     */
    /* attributes to the "output-link" attribute on the object  */
    /* which is value of the "io" attribute on the top-state.   */
    /* For example, if a Soar production adds the attribute     */
    /* "command" with the attribute "square-result" with value  */
    /* 9, the resulting structure would look something like     */
    /* this:                                                    */
    /*                                                          */
    /* S1 ^io I1                                                */
    /*    I1 ^output-link O37                                   */
    /*       O37 ^command print-square-result                   */
    /*           ^square-result 9                               */
    /*                                                          */
    
    add_output_function(a, my_output_function, NULL, NULL, "output-link"); 

    break;
  }
}


